#!/usr/bin/python
#	Copyright (C) 2007 Ghislain LEVEQUE <ghislain.leveque@gmail.com>
#
#	This file is part of NetBlocks.
#	
#	NetBlocks is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#	
#	NetBlocks is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#	
#	You should have received a copy of the GNU General Public License
#	along with NetBlocks; if not, write to the Free Software
#	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# init {{{1
# imports {{{2

import sys,os,random
import platform
import pygame
import pygame.font
from pygame.locals import *

#constants {{{2
PATHNAME=os.path.dirname(sys.argv[0])
FULLPATH=os.path.abspath(PATHNAME)
DATAPATH=os.path.join(FULLPATH,"data/")

SCREEN_WIDTH=800
SCREEN_HEIGHT=600
CELL_SIZE=32
GAME_BOARD_POS=(240,90)

MAIN_GAME_NAME="NetBlocks"
GAME_VERSION="alpha"
GAME_NAME=MAIN_GAME_NAME+" "+GAME_VERSION
SYMBOLS=[1 | 4,
		2 | 8,
		1 | 2 | 4,
		1 | 4 | 8,
		1 | 2 | 8,
		2 | 4 | 8,
		2 | 4,
		4 | 8,
		1 | 8,
		1 | 2,
		]
# a level consist of a wave speed (in ms), a minimum of active cells for deletion and the score needed to reach the next level:
# if no level definition is present here, the game will take the last active level, substract 500 from the wave speed, add 0.5 to minimum active cells and add 50 to next level score
LEVELS=[
		[10000,0,10],
		[10000,1,20],
		[9500,2,35],
		[9500,3,50],
		[9000,4,65],
		[9000,5,85],
		[8500,6,105],
		[8500,7,125],
		[8000,8,145],
		[8000,9,175],
		[7500,10,205]
		]

# this formula is used to calculate the real score depending on the number of cells cleared at once
def FORMULE_SCORE(value):
	return value*value-3*value+4

SYMBOL_TOPWIRE=1
SYMBOL_RIGHTWIRE=2
SYMBOL_BOTTOMWIRE=4
SYMBOL_LEFTWIRE=8
SYMBOL_CLIENT=16

BLACK=(0,0,0)
RED=(255,0,0)
GREEN=(0,255,0)
BLUE=(0,0,255)

TRUE=1
FALSE=0


# Classes {{{1
# Events {{{2
class Event: #{{{3
	def __init__(self):
		self.name="Generic Event"

class MenuSelectedEvent(Event): #{{{3
	def __init__(self):
		self.name="A Menu was selected, event"

class WaveArrivedEvent(Event): #{{{3
	def __init__(self):
		self.name="a wave arrived"

class GameOverEvent(Event): #{{{3
	def __init__(self):
		self.name="the game is over"

class GroupOfCellClearedEvent(Event): #{{{3
	def __init__(self):
		self.name="A group of cell was cleared"

class ToggleFullscreen(Event): #{{{3
	def __init__(self):
		self.name="Toggle full screen event"

class SetLevelEvent(Event): #{{{3
	def __init__(self,level,nlevel,mincells):
		self.level=level
		self.nlevel=nlevel
		self.mincells=mincells
		self.name="Set level event : "+str(self.level)

class UpdateScoreEvent(Event): #{{{3
	def __init__(self,score,logscore):
		self.score=score
		self.logscore=logscore
		self.name="Update score event : "+str(self.score)+" ("+str(self.logscore)+")"

class PlayerWantsWave(Event): #{{{3
	def __init__(self):
		self.name="Player wants a wave event"


class SelectEvent(Event): #{{{3
	def __init__(self,pos,button):
		self.name="Mouse selection"
		self.pos=pos
		self.button=button

class CellClickedEvent(Event): #{{{3
	def __init__(self,cell,button):
		self.button=button
		self.cell=cell
		self.name="Cell clicked at ["+str(cell.x)+","+str(cell.y)+"] with button "+str(self.button)

class ClientActivationEvent(Event): #{{{3
	def __init__(self,cell):
		self.cell=cell
		self.name="Client activation at ["+str(cell.x)+","+str(cell.y)+"]"

class TickEvent(Event):  #{{{3
	def __init__(self):
		self.name = "CPU Tick Event"

class QuitEvent(Event):  #{{{3
	def __init__(self):
		self.name = "Program Quit Event"


class NewGameEvent(Event):  #{{{3
	def __init__(self):
		self.name = "New Game Event"


class GameReadyEvent(Event):  #{{{3
	def __init__(self):
		self.name = "Game Ready Event"


# Event and Spinner Managment {{{2
class EventManager:  #{{{3
	"""cet objet est responsable de la coordination des communications
	entre le Modele, la Vue et le Controlleur."""
	def __init__(self ):
		from weakref import WeakKeyDictionary
		self.listeners = WeakKeyDictionary()
		self.eventQueue= []

	#----------------------------------------------------------------------
	def RegisterListener( self, listener ):
		self.listeners[ listener ] = 1

	#----------------------------------------------------------------------
	def UnregisterListener( self, listener ):
		if listener in self.listeners.keys():
			del self.listeners[ listener ]
		
	#----------------------------------------------------------------------
	def Post( self, events ):
		for e in events:
			if not isinstance(e, TickEvent): print( "     Message: " + e.name )
		for listener in self.listeners.keys():
			#NOTE: If the weakref has died, it will be 
			#automatically removed, so we don't have 
			#to worry about it.
                        for e in events:
                            listener.Notify( e )

class Controller:  #{{{3
	"""..."""
	def __init__(self, evManager):
		self.evManager = evManager
		self.evManager.RegisterListener( self )

	#----------------------------------------------------------------------
	def Notify(self, event):
		if isinstance( event, TickEvent ):
			#Handle Input Events
			for event in pygame.event.get():
				ev = None
				if event.type == QUIT:
					ev = QuitEvent()
				elif event.type == KEYDOWN \
                                     and event.key == K_f:
                                        ev = ToggleFullscreen()
				elif event.type == KEYDOWN \
                                     and event.key == K_SPACE:
                                        ev = PlayerWantsWave()
				elif event.type == KEYDOWN \
				     and event.key == K_ESCAPE:
					ev = QuitEvent()
				elif event.type == MOUSEBUTTONDOWN:
					ev = SelectEvent(event.pos,event.button)
				if ev:
					self.evManager.Post( [ev] )

class CPUSpinnerController:  #{{{3
	"""..."""
	def __init__(self, evManager):
		self.evManager = evManager
		self.evManager.RegisterListener( self )

		self.keepGoing = 1

	#----------------------------------------------------------------------
	def Run(self):
		while self.keepGoing:
			event = TickEvent()
			self.evManager.Post( [event] )

	#----------------------------------------------------------------------
	def Notify(self, event):
		if isinstance( event, QuitEvent ):
			#this will stop the while loop from running
			self.keepGoing = 0
			sys.exit(0)

# Display and sound management {{{2
# Sprites {{{3
class CellSprite(pygame.sprite.Sprite): #{{{4
	def __init__(self, group=None):
	       pygame.sprite.Sprite.__init__(self, group)
	       Surface = pygame.Surface( (CELL_SIZE,CELL_SIZE) )
	       self.image = Surface
	       self.rect  = Surface.get_rect()

	def DrawSymbol(self,symbol,active,locked,todelete):
		Surface=self.image
	        pygame.draw.rect(Surface,BLACK,pygame.Rect((0,0),(CELL_SIZE,CELL_SIZE)))
	        pygame.draw.rect(Surface,(255,255,0),pygame.Rect((2,2),(CELL_SIZE-2,CELL_SIZE-2)))
		if active:
			color=BLUE
		elif locked:
			color=GREEN
		elif todelete:
			color=RED
		else:
			color=BLACK
		pygame.draw.rect(Surface,(255,255,0),pygame.Rect((2,2),(CELL_SIZE-2,CELL_SIZE-2)))
		if symbol & SYMBOL_TOPWIRE:
			pygame.draw.line(Surface,color,(CELL_SIZE/2,CELL_SIZE/2),(CELL_SIZE/2,0),3)
		if symbol & SYMBOL_RIGHTWIRE:
			pygame.draw.line(Surface,color,(CELL_SIZE/2,CELL_SIZE/2),(CELL_SIZE,CELL_SIZE/2),3)
		if symbol & SYMBOL_BOTTOMWIRE:
			pygame.draw.line(Surface,color,(CELL_SIZE/2,CELL_SIZE/2),(CELL_SIZE/2,CELL_SIZE),3)
		if symbol & SYMBOL_LEFTWIRE:
			pygame.draw.line(Surface,color,(CELL_SIZE/2,CELL_SIZE/2),(0,CELL_SIZE/2),3)
		if symbol & SYMBOL_CLIENT:
			pygame.draw.rect(Surface,color,pygame.Rect((CELL_SIZE/2-4,CELL_SIZE/2-4,8,8)))

class TextSprite(pygame.sprite.Sprite): #{{{4
	def __init__(self,group=None,text="Text"):
		self.text=text
		self.score=0
		pygame.sprite.Sprite.__init__(self,group)
		Surface=pygame.image.load(DATAPATH+"button.png").convert_alpha()
		self.image=Surface
		self.event=Event()
		self.rect=Surface.get_rect()

	def Render(self):
		self.image=pygame.image.load(DATAPATH+"button.png").convert_alpha()
		ZeSize=54
		ZeFont=pygame.font.Font(None,ZeSize)
		while (ZeFont.size(self.text+str(self.score))[0]>self.image.get_rect().width):
			ZeSize-=1
			if (ZeSize==1):
				break
			ZeFont=pygame.font.Font(None,ZeSize)
		ZeText=ZeFont.render(self.text+str(self.score),1,(255,0,0))
		self.image.blit(ZeText,(
			((self.image.get_rect().width-ZeFont.size(self.text+str(self.score))[0])/2),
			((self.image.get_rect().height-ZeFont.size(self.text+str(self.score))[1])/2))
			)
		ZeFont=pygame.font.Font(None,54)
		ZeText=ZeFont.render(self.text+str(self.score),1,GREEN)

	def Update(self,value):
		self.score=value
		self.Render()

class BackSprite(pygame.sprite.Sprite): #{{{4
	def __init__(self,group=None):
		pygame.sprite.Sprite.__init__(self,group)
		Surface=pygame.image.load(DATAPATH+"background.png").convert_alpha()
		self.image=Surface
		self.rect=Surface.get_rect()

class ButtonSprite(pygame.sprite.Sprite): #{{{4
	def __init__(self,group=None,text="Bouton"):
		pygame.sprite.Sprite.__init__(self,group)
		Surface=pygame.image.load(DATAPATH+"button.png").convert_alpha()
		ZeSize=54
		ZeFont=pygame.font.Font(None,ZeSize)
		while (ZeFont.size(text)[0]>Surface.get_rect().width):
			ZeSize-=1
			if (ZeSize==1):
				break
			ZeFont=pygame.font.Font(None,ZeSize)
		ZeText=ZeFont.render(text,1,(255,0,0))
		Surface.blit(ZeText,(
			((Surface.get_rect().width-ZeFont.size(text)[0])/2),
			((Surface.get_rect().height-ZeFont.size(text)[1])/2))
			)
		self.image=Surface
		self.rect=Surface.get_rect()
		self.event=Event()

	def bind(self,event):
		self.event=event

class PyPartieView:  #{{{3
	"""..."""
	def __init__(self, evManager, partie):
		self.evManager = evManager
		self.evManager.RegisterListener( self )
		self.partie = partie

		pygame.init()
		self.window = pygame.display.set_mode( (SCREEN_WIDTH,SCREEN_HEIGHT) )
		pygame.display.set_caption(GAME_NAME )
		self.background = pygame.Surface( self.window.get_size() )
		pygame.display.update(self.background.get_rect())

		self.backSprites = pygame.sprite.OrderedUpdates()
		self.boutonsSprites = pygame.sprite.OrderedUpdates()
		self.frontSprites = pygame.sprite.OrderedUpdates()
		Test=BackSprite(self.backSprites)
		Test.rect=self.background.get_rect()
		self.logscoreSprite=TextSprite(self.boutonsSprites,"Score : ")
		self.logscoreSprite.rect.center=(SCREEN_WIDTH-100,100)
		self.scoreSprite=TextSprite(self.boutonsSprites,"Cells : ")
		self.scoreSprite.rect.center=(SCREEN_WIDTH-100,150)
		self.levelSprite=TextSprite(self.boutonsSprites,"Level : ")
		self.levelSprite.rect.center=(SCREEN_WIDTH-100,300)
		self.nextlevelSprite=TextSprite(self.boutonsSprites,"Next : ")
		self.nextlevelSprite.rect.center=(SCREEN_WIDTH-100,400)
		self.mincellsSprite=TextSprite(self.boutonsSprites,"Min. : ")
		self.mincellsSprite.rect.center=(SCREEN_WIDTH-100,500)
		BoutonTest=ButtonSprite(self.boutonsSprites,"New Game")
		BoutonTest.rect.center=(100,100)
		BoutonTest.bind(NewGameEvent())
		BoutonTest=ButtonSprite(self.boutonsSprites,"Quit Game")
		BoutonTest.rect.center=(100,500)
		BoutonTest.bind(QuitEvent())

	def createCellsSprites(self):
		cells=self.partie.cells
		for col in cells:
			for cell in col:
				cell.AssignSprite(CellSprite(self.frontSprites))

	def Notify(self, event):
		if isinstance( event, TickEvent ):
			#Draw Everything
			self.backSprites.clear( self.window, self.background )
			self.boutonsSprites.clear( self.window, self.background )
			self.frontSprites.clear( self.window, self.background )

			self.backSprites.update()
			self.boutonsSprites.update()
			self.frontSprites.update()

			dirtyRects1 = self.backSprites.draw( self.window )
			dirtyRects2 = self.boutonsSprites.draw( self.window )
			dirtyRects3 = self.frontSprites.draw( self.window )
			
			dirtyRects = dirtyRects1 + dirtyRects2 + dirtyRects3
			pygame.display.update( dirtyRects )
		elif isinstance(event, UpdateScoreEvent):
			self.scoreSprite.Update(event.score)
			self.logscoreSprite.Update(event.logscore)
		elif isinstance(event, SetLevelEvent):
			self.levelSprite.Update(event.level)
			self.nextlevelSprite.Update(event.nlevel)
			self.mincellsSprite.Update(event.mincells)
		elif isinstance(event, ToggleFullscreen):
                        pygame.display.toggle_fullscreen()
		elif isinstance(event,GameReadyEvent):
			self.scoreSprite.Update(0)
			self.logscoreSprite.Update(0)
			self.levelSprite.Update(0)
			self.createCellsSprites()


class PyGameSound: #{{{3
	def __init__(self,evManager):
		self.evManager=evManager
		self.evManager.RegisterListener(self)
		if pygame.mixer:
			pygame.mixer.init
			print "pygame.mixer initialized : ",pygame.mixer.get_num_channels(),"channels available"
			self.downSound=pygame.mixer.Sound(DATAPATH+"down.ogg")
			self.clicSound=pygame.mixer.Sound(DATAPATH+"clic.ogg")
			self.plopSound=pygame.mixer.Sound(DATAPATH+"plop.ogg")
			self.warpSound=pygame.mixer.Sound(DATAPATH+"warp.ogg")
			self.nlevelSound=pygame.mixer.Sound(DATAPATH+"nextlevel.ogg")
			self.downSound.play()

	def Notify(self,event):
		if isinstance(event,MenuSelectedEvent) or isinstance(event,GroupOfCellClearedEvent):
			self.plopSound.play()
		elif isinstance(event,CellClickedEvent):
			self.clicSound.play()
		elif isinstance(event,WaveArrivedEvent):
			self.warpSound.play()
		elif isinstance(event,GameOverEvent):
			self.downSound.play()
		elif isinstance(event,SetLevelEvent):
			self.nlevelSound.play()

# Main Game engine {{{2
class Cell: #{{{3
	def __init__(self,x,y):
		self.x=x
		self.y=y
		self.sprite=None
		self.symbol=0
		self.active=FALSE
		self.locked=FALSE
		self.todelete=FALSE

	def AssignSymbol(self,symbol):
		self.symbol=symbol
		self.Update()

	def AssignSprite(self,sprite):
		self.sprite=sprite
		self.sprite.rect.center=((self.x+0.5)*CELL_SIZE+GAME_BOARD_POS[0],(self.y+0.5)*CELL_SIZE+GAME_BOARD_POS[1])
		self.Update()

	def Rotate(self,counter=None):
		if self.locked: return
		wires=self.symbol & 15
		client=self.symbol & 16
		if not counter:
			one=wires & SYMBOL_TOPWIRE
			wires=wires >> 1
			if one: wires = wires | SYMBOL_LEFTWIRE
		else:
			eight=wires & SYMBOL_LEFTWIRE
			wires=wires << 1
			if eight: wires=wires | SYMBOL_TOPWIRE
		wires=wires & 15
		self.symbol=wires
		if client: self.symbol = self.symbol | SYMBOL_CLIENT
		self.Update()

	def PutAClient(self):
		self.symbol=self.symbol | SYMBOL_CLIENT
		self.Update()

	def Activate(self):
		self.active=TRUE
		self.Update()

	def Lock(self):
		self.locked=TRUE
		self.Update()

	def MarkToDelete(self):
		self.todelete=TRUE
		self.Update()

	def Clear(self):
		self.symbol=0
		self.active=FALSE
		self.locked=FALSE
		self.todelete=FALSE
		self.Update()

	def CopyFromCell(self,cell):
		self.symbol=cell.symbol
		self.active=cell.active
		self.locked=cell.locked
		self.todelete=cell.todelete
		self.Update()

	def Update(self):
		self.sprite.DrawSymbol(self.symbol,self.active,self.locked,self.todelete)

class Partie: #{{{3
	def __init__(self,evManager): #{{{4
		self.evManager=evManager
		self.evManager.RegisterListener(self)
		self.reset()

	def reset(self): #{{{4
		self.cells=[]
		self.nextwave_time=0
		self.current_speed=LEVELS[0][0]
		self.minimum_active_cells=LEVELS[0][1]
		self.next_level_score=LEVELS[0][2]
		self.state=0
		self.score=0
		self.logscore=0
		self.level=0

	def prepare(self): #{{{4
		self.reset()
		for i in range(0,10):
			col=[]
			for j in range (0,15):
				col.append(Cell(i,j))
			self.cells.append(col)
		self.state=1
		self.evManager.Post([SetLevelEvent(self.level,self.next_level_score,int(self.minimum_active_cells)),GameReadyEvent()])

	def SendWave(self): #{{{4
		if not self.state: return
		for col in self.cells:
			for cell in col:
				if cell.symbol==0:
					cell.AssignSymbol(SYMBOLS[random.randrange(0,10)])
					if random.randrange(0,5)==1:
						cell.PutAClient()
						self.ActivateCellsFrom(cell)
					break
			if cell.y==14:
				print "game over"
				self.state=0
				self.evManager.Post([GameOverEvent()])
		self.evManager.Post([WaveArrivedEvent()])
		self.nextwave_time=pygame.time.get_ticks()+self.current_speed
		self.UpdateStatus()
	
	def Gravity(self): #{{{4
		mouvment=1
		while mouvment:
			mouvment=0
			for col in self.cells:
				for cell in col:
					if cell.y>0 and cell.symbol!=0:
						topcell=col[cell.y-1]
						if topcell.symbol==0:
							mouvment=1
							col[cell.y-1].CopyFromCell(cell)
							cell.Clear()

	def ActivateCellsFrom(self,cell): #{{{4
		#print "activating cell at",cell.x,cell.y
		cell.Activate()
		if cell.symbol & SYMBOL_TOPWIRE:
			#print cell.x,cell.y,"it has a top wire"
			if cell.y>0:
				neighbour=self.cells[cell.x][cell.y-1]
				if neighbour.symbol & SYMBOL_BOTTOMWIRE and not neighbour.active:
					#print "cell at ",neighbour.x,neighbour.y,"has a bottom wire and is inactive"
					self.ActivateCellsFrom(neighbour)
					#print "neighbour",neighbour.x,neighbour.y,"has been activated"
		if cell.symbol & SYMBOL_RIGHTWIRE:
			#print cell.x,cell.y,"it has a right wire"
			if cell.x<9:
				neighbour=self.cells[cell.x+1][cell.y]
				if neighbour.symbol & SYMBOL_LEFTWIRE and not neighbour.active:
					#print "cell at ",neighbour.x,neighbour.y,"has a left wire and is inactive"
					self.ActivateCellsFrom(neighbour)
					#print "neighbour",neighbour.x,neighbour.y,"has been activated"
		if cell.symbol & SYMBOL_BOTTOMWIRE:
			#print cell.x,cell.y,"it has a bottom wire"
			if cell.y<14:
				neighbour=self.cells[cell.x][cell.y+1]
				if neighbour.symbol & SYMBOL_TOPWIRE and not neighbour.active:
					#print "cell at ",neighbour.x,neighbour.y,"has a top wire and is inactive"
					self.ActivateCellsFrom(neighbour)
					#print "neighbour",neighbour.x,neighbour.y,"has been activated"
		if cell.symbol & SYMBOL_LEFTWIRE:
			#print cell.x,cell.y,"it has a left wire"
			if cell.x>0:
				neighbour=self.cells[cell.x-1][cell.y]
				if neighbour.symbol & SYMBOL_RIGHTWIRE and not neighbour.active:
					#print "cell at ",neighbour.x,neighbour.y,"has a right wire and is inactive"
					self.ActivateCellsFrom(neighbour)
					#print "neighbour",neighbour.x,neighbour.y,"has been activated"

	def ClearActiveCellsFrom(self,cell): #{{{4
		#print "lockating cell at",cell.x,cell.y
		cell.Lock()
		how_many_cells_locked=1
		if cell.symbol & SYMBOL_TOPWIRE:
			#print cell.x,cell.y,"it has a top wire"
			if cell.y>0:
				neighbour=self.cells[cell.x][cell.y-1]
				if neighbour.symbol & SYMBOL_BOTTOMWIRE and not neighbour.locked:
					#print "cell at ",neighbour.x,neighbour.y,"has a bottom wire and is inlocke"
					how_many_cells_locked+=self.ClearActiveCellsFrom(neighbour)
					#print "neighbour",neighbour.x,neighbour.y,"has been lockated"
		if cell.symbol & SYMBOL_RIGHTWIRE:
			#print cell.x,cell.y,"it has a right wire"
			if cell.x<9:
				neighbour=self.cells[cell.x+1][cell.y]
				if neighbour.symbol & SYMBOL_LEFTWIRE and not neighbour.locked:
					#print "cell at ",neighbour.x,neighbour.y,"has a left wire and is inlock"
					how_many_cells_locked+=self.ClearActiveCellsFrom(neighbour)
					#print "neighbour",neighbour.x,neighbour.y,"has been lockated"
		if cell.symbol & SYMBOL_BOTTOMWIRE:
			#print cell.x,cell.y,"it has a bottom wire"
			if cell.y<14:
				neighbour=self.cells[cell.x][cell.y+1]
				if neighbour.symbol & SYMBOL_TOPWIRE and not neighbour.locked:
					#print "cell at ",neighbour.x,neighbour.y,"has a top wire and is inlock"
					how_many_cells_locked+=self.ClearActiveCellsFrom(neighbour)
					#print "neighbour",neighbour.x,neighbour.y,"has been lockated"
		if cell.symbol & SYMBOL_LEFTWIRE:
			#print cell.x,cell.y,"it has a left wire"
			if cell.x>0:
				neighbour=self.cells[cell.x-1][cell.y]
				if neighbour.symbol & SYMBOL_RIGHTWIRE and not neighbour.locked:
					#print "cell at ",neighbour.x,neighbour.y,"has a right wire and is inlock"
					how_many_cells_locked+=self.ClearActiveCellsFrom(neighbour)
					#print "neighbour",neighbour.x,neighbour.y,"has been lockated"
		return how_many_cells_locked

	def DeleteLockedCells(self): #{{{4
		score=0
		for col in self.cells:
			for cell in col:
				if cell.locked:
					cell.Clear()
					cell.Update()
					score+=1
		return score

	def ClearCellsActivation(self): #{{{4
		for col in self.cells:
			for cell in col:
				if cell.active:
					cell.active=FALSE
					cell.Update()

	def UnLockCells(self): #{{{4
		for col in self.cells:
			for cell in col:
				if cell.locked:
					cell.locked=FALSE
					cell.Update()

	def UpdateStatus(self): #{{{4
		self.Gravity()
		self.ClearCellsActivation()
		for col in self.cells:
			for cell in col:
				if cell.symbol & SYMBOL_CLIENT:
					self.ActivateCellsFrom(cell)

	def NextLevel(self): #{{{4
		self.level+=1
		if self.level<len(LEVELS):
			self.current_speed=LEVELS[self.level][0]
			self.minimum_active_cells=LEVELS[self.level][1]
			self.next_level_score=LEVELS[self.level][2]
		else:
			self.current_speed-=500
			self.minimum_active_cells+=1
			self.next_level_score+=50
		print "new level, speed",self.current_speed,"minactcell",self.minimum_active_cells,"nextlevelscore",self.next_level_score
		self.evManager.Post([SetLevelEvent(self.level,self.next_level_score,int(self.minimum_active_cells))])

	def AddToScore(self,value): #{{{4
		self.score+=value
		self.logscore+=FORMULE_SCORE(value)
		while self.score>self.next_level_score:
			self.NextLevel()
		self.evManager.Post([UpdateScoreEvent(self.score,self.logscore)])

	def Notify(self,event): #{{{4
		ev=None
		if isinstance(event,NewGameEvent):
			self.prepare()
		elif self.state and isinstance(event,CellClickedEvent) and event.button!=2:
			event.cell.Rotate((event.button==3))
			self.UpdateStatus()
		elif self.state and isinstance(event,ClientActivationEvent):
			nbCellsLocked=self.ClearActiveCellsFrom(event.cell)
			if nbCellsLocked>self.minimum_active_cells:
				self.AddToScore(self.DeleteLockedCells())
				ev=GroupOfCellClearedEvent()
			else:
				self.UnLockCells()
			self.UpdateStatus()
		elif self.state and isinstance(event,PlayerWantsWave):
			self.SendWave()
		elif self.state and isinstance(event,TickEvent):
			time=pygame.time.get_ticks()
			if time>self.nextwave_time:
				self.SendWave()
		if ev:
			self.evManager.Post([ev])

class Selector: #{{{3
	def __init__(self,evManager,pygameView):
		self.evManager=evManager
		self.associatedView=pygameView
		self.evManager.RegisterListener(self)

	def IsAButton(self,event):
		for sprite in self.associatedView.boutonsSprites:
			if sprite.rect.collidepoint(event.pos):
				return sprite.event
		return None

	def IsACell(self,event):
		for col in self.associatedView.partie.cells:
			for cell in col:
				if cell.sprite.rect.collidepoint(event.pos):
					if cell.active and event.button==2:
						return ClientActivationEvent(cell)
					return CellClickedEvent(cell,event.button)
		return None

	def WhatIsSelected(self,event):
		events=[]
		isbutton=self.IsAButton(event)
		if isbutton:
			events.append(isbutton)
			events.append(MenuSelectedEvent())
		iscell=self.IsACell(event)
		if iscell:
			events.append(iscell)
		return events

	def Notify(self,event):
		events=[]
		if isinstance(event,SelectEvent):
			events=self.WhatIsSelected(event)
		if events:
			self.evManager.Post(events)


# main {{{1

def main():
	print "Welcome to",GAME_NAME
	evManager = EventManager()

	contrl = Controller( evManager )
	spinner = CPUSpinnerController( evManager )
	partie = Partie( evManager )
	pygameView = PyPartieView( evManager, partie )
	soundCtrler = PyGameSound(evManager)
	selector= Selector(evManager,pygameView)
	
	spinner.Run()

if __name__=="__main__":
	main()
